SELECT
  RANK() OVER (ORDER BY "t0"."double_col" ASC ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) - 1 AS "rank",
  DENSE_RANK() OVER (ORDER BY "t0"."double_col" ASC ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) - 1 AS "dense_rank",
  CUME_DIST() OVER (ORDER BY "t0"."double_col" ASC ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) AS "cume_dist",
  NTILE(7) OVER (ORDER BY "t0"."double_col" ASC ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) - 1 AS "ntile",
  PERCENT_RANK() OVER (ORDER BY "t0"."double_col" ASC ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) AS "percent_rank"
FROM "functional_alltypes" AS "t0"