SELECT t0.*,
       avg(t0.`float_col`) OVER (ORDER BY UNIX_MICROS(t0.`timestamp_col`) ASC RANGE BETWEEN 31536000000000 PRECEDING AND EXTRACT(YEAR from INTERVAL 0 YEAR) * 31536000000000 FOLLOWING) AS `win_avg`
FROM functional_alltypes t0